#include <stdio.h>
#include <assert.h>
#include "app.h"
#include "opengl.h"
#include "sdr.h"
#include "texture.h"
#include "mesh.h"
#include "meshgen.h"
#include "scene.h"
#include "metascene.h"
#include "datamap.h"

static void draw_scene();

long time_msec;
int win_width, win_height;
bool opt_gear_wireframe;

static float cam_dist = 0.0;
static float cam_theta, cam_phi = 20;
static Vec3 cam_pos;
static int prev_mx, prev_my;
static bool bnstate[8];
static bool keystate[256];

static Mat4 view_matrix;
static TextureSet texman;
static Scene *scn;
static unsigned int sdr;

static long prev_msec;


bool app_init()
{
	glEnable(GL_FRAMEBUFFER_SRGB);
	glEnable(GL_MULTISAMPLE);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glEnable(GL_LIGHTING);
	glEnable(GL_NORMALIZE);

	Mesh::use_custom_sdr_attr = false;

	float ambient[] = {0.0, 0.0, 0.0, 0.0};
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);

	scn = new Scene(&texman);
	if(!load_scene(scn, "data/museum.scene")) {
		return false;
	}

	/*
	datamap_set_path("data");
	if(!datamap_load_map("data.map")) {
		fprintf(stderr, "failed to load datafile mappings\n");
	}

	unsigned int sflags = SCNLOAD_FLIPTEX;
	if(!(scn->load("data/testscene/patoma.fbx", sflags)) ||
			!(scn->load("data/testscene/kolones.fbx", sflags))) {
		fprintf(stderr, "failed to load test scene\n");
		return false;
	}
	*/

	if(!(sdr = create_program_load("sdr/test.v.glsl", "sdr/test.p.glsl"))) {
		fprintf(stderr, "failed to load test shaders\n");
		return false;
	}
	set_uniform_int(sdr, "texmap", 0);
	set_uniform_int(sdr, "lightmap", 1);

	glUseProgram(0);
	return true;
}

void app_cleanup()
{
	texman.clear();
}

static void update(float dt)
{
	texman.update();

	scn->update(dt);

	float walk_speed = 2000.0 * dt;
	Vec3 dir;

	if(keystate[(int)'w']) {
		dir.z -= walk_speed;
	}
	if(keystate[(int)'s']) {
		dir.z += walk_speed;
	}
	if(keystate[(int)'d']) {
		dir.x += walk_speed;
	}
	if(keystate[(int)'a']) {
		dir.x -= walk_speed;
	}
	if(keystate[(int)'q']) {
		cam_pos.y += walk_speed;
	}
	if(keystate[(int)'z']) {
		cam_pos.y -= walk_speed;
	}

	float theta = M_PI * cam_theta / 180.0f;
	cam_pos.x += cos(theta) * dir.x - sin(theta) * dir.z;
	cam_pos.z += sin(theta) * dir.x + cos(theta) * dir.z;
}

static void set_light(int idx, const Vec3 &pos, const Vec3 &color)
{
	unsigned int lt = GL_LIGHT0 + idx;
	float posv[] = { pos.x, pos.y, pos.z, 1 };
	float colv[] = { color.x, color.y, color.z, 1 };

	glEnable(lt);
	glLightfv(lt, GL_POSITION, posv);
	glLightfv(lt, GL_DIFFUSE, colv);
	glLightfv(lt, GL_SPECULAR, colv);
}

void app_display()
{
	float dt = (float)(time_msec - prev_msec) / 1000.0f;
	prev_msec = time_msec;

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	view_matrix = Mat4::identity;
	view_matrix.pre_translate(0, 0, -cam_dist);
	view_matrix.pre_rotate(deg_to_rad(cam_phi), 1, 0, 0);
	view_matrix.pre_rotate(deg_to_rad(cam_theta), 0, 1, 0);
	view_matrix.pre_translate(-cam_pos.x, -cam_pos.y, -cam_pos.z);

	glMatrixMode(GL_MODELVIEW);
	glLoadMatrixf(view_matrix[0]);

	static const Vec3 lpos[] = { Vec3(-50, 75, 100), Vec3(100, 0, 30), Vec3(-10, -10, 60) };
	set_light(0, lpos[0], Vec3(1.0, 0.8, 0.7) * 0.8);
	set_light(1, lpos[1], Vec3(0.6, 0.7, 1.0) * 0.6);
	set_light(2, lpos[2], Vec3(0.8, 1.0, 0.8) * 0.3);

	update(dt);

	draw_scene();

	app_swap_buffers();
	assert(glGetError() == GL_NO_ERROR);
}


static void draw_scene()
{
	/*
	glBegin(GL_QUADS);
	glNormal3f(0, 1, 0);
	glVertex3f(-30, -10, 30);
	glVertex3f(30, -10, 30);
	glVertex3f(30, -10, -30);
	glVertex3f(-30, -10, -30);
	glEnd();
	*/

	glUseProgram(sdr);
	scn->draw();
	glUseProgram(0);
}


void app_reshape(int x, int y)
{
	glViewport(0, 0, x, y);

	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(50.0, (float)x / (float)y, 1.0, 10000.0);
}

void app_keyboard(int key, bool pressed)
{
	if(pressed) {
		switch(key) {
		case 27:
			app_quit();
			break;
		}
	}

	keystate[key] = pressed;
}

void app_mouse_button(int bn, bool pressed, int x, int y)
{
	prev_mx = x;
	prev_my = y;
	bnstate[bn] = pressed;
}

void app_mouse_motion(int x, int y)
{
	int dx = x - prev_mx;
	int dy = y - prev_my;
	prev_mx = x;
	prev_my = y;

	if(!dx && !dy) return;

	if(bnstate[0]) {
		cam_theta += dx * 0.5;
		cam_phi += dy * 0.5;

		if(cam_phi < -90) cam_phi = -90;
		if(cam_phi > 90) cam_phi = 90;
	}
	if(bnstate[2]) {
		cam_dist += dy * 0.1;
		if(cam_dist < 0.0) cam_dist = 0.0;
	}
}
