#include <stdio.h>
#include <assert.h>
#include <vector>
#include <map>
#include <gmath/gmath.h>
#include <assimp/cimport.h>
#include <assimp/postprocess.h>
#include <assimp/scene.h>
#include <assimp/mesh.h>
#include <assimp/material.h>
#include <assimp/anim.h>
#include <assimp/vector3.h>
#include <assimp/matrix4x4.h>
#include <assimp/quaternion.h>
#include "scene.h"
#include "objmesh.h"
#include "datamap.h"

static bool load_material(Scene *scn, Material *mat, const aiMaterial *aimat);
static SceneNode *load_node(Scene *scn, const aiScene *aiscn, unsigned int flags, const aiNode *ainode);
static Mesh *load_mesh(Scene *scn, const aiScene *aiscn, unsigned int flags, const aiMesh *aimesh);
static void print_nodes(SceneNode *node, int lvl = 0);
/*static const char *mprop_semantic(int x);
static int count_textures(const aiMaterial *aimat);*/
static int assimp_textype(aiTextureType type);
static const char *assimp_textypestr(aiTextureType type);

static Mat4 assimp_matrix(const aiMatrix4x4 &aim);

/*static Vec3 assimp_vector(const aiVector3D &v);
static Quat assimp_quat(const aiQuaternion &q);
static long assimp_time(const aiAnimation *anim, double aitime);
static void print_hierarchy(const aiNode *node);
*/

static std::map<std::string, SceneNode*> node_by_name;
static std::map<aiMesh*, Mesh*> mesh_by_aimesh;

bool Scene::load(const char *fname, unsigned int flags)
{
	unsigned int ppflags = aiProcess_CalcTangentSpace |
		aiProcess_GenNormals |
		aiProcess_JoinIdenticalVertices |
		aiProcess_Triangulate |
		aiProcess_SortByPType |
		aiProcess_GenUVCoords |
		//aiProcess_PreTransformVertices |
		aiProcess_TransformUVCoords;

	if(flags & SCNLOAD_FLIPTEX) {
		ppflags |= aiProcess_FlipUVs;
	}

	printf("Loading scene file: %s\n", fname);

	const aiScene *aiscn = aiImportFile(fname, ppflags);
	if(!aiscn) {
		fprintf(stderr, "failed to load scene file: %s\n", fname);
		return false;
	}

	// assimp adds its own root node, which might have transformations
	Vec3 root_pos, root_scaling(1.0, 1.0, 1.0);
	Quat root_rot;

	if(aiscn->mRootNode) {
		Mat4 root_matrix = assimp_matrix(aiscn->mRootNode->mTransformation);
		root_pos = root_matrix.get_translation();
		root_rot = root_matrix.get_rotation();
		root_scaling = root_matrix.get_scaling();

		printf("assimp root node: %s\n", aiscn->mRootNode->mName.data);
		printf("  pos: %f %f %f\n", root_pos.x, root_pos.y, root_pos.z);
		printf("  rot: %f %+f %+f %+f\n", root_rot.w, root_rot.x, root_rot.y, root_rot.z);
		printf("  scaling: %f %f %f\n", root_scaling.x, root_scaling.y, root_scaling.z);
	}

	// load all meshes
	for(unsigned int i=0; i<aiscn->mNumMeshes; i++) {
		aiMesh *aimesh = aiscn->mMeshes[i];
		Mesh *mesh;

		switch(aimesh->mPrimitiveTypes) {
		case aiPrimitiveType_TRIANGLE:
			if((mesh = load_mesh(this, aiscn, flags, aimesh))) {
				mesh_by_aimesh[aimesh] = mesh;
				meshes.push_back(mesh);
			}
			break;

		default:
			fprintf(stderr, "unsupported primitive type: %u\n", aimesh->mPrimitiveTypes);
			break;
		}
	}

	if(!nodes) {
		nodes = new SceneNode;
		nodes->set_name("root");
		nodes->set_position(root_pos);
		nodes->set_rotation(root_rot);
		nodes->set_scaling(root_scaling);
	}

	// load all the nodes recursively
	for(unsigned int i=0; i<aiscn->mRootNode->mNumChildren; i++) {
		SceneNode *node = load_node(this, aiscn, flags, aiscn->mRootNode->mChildren[i]);
		if(node) {
			nodes->add_child(node);
		}
	}

	node_by_name.clear();
	mesh_by_aimesh.clear();

	aiReleaseImport(aiscn);
	printf("loaded scene file: %s, %d meshes\n", fname, (int)meshes.size());
	nodes->update(0);
	print_nodes(nodes);
	return true;
}

static bool load_material(Scene *scn, Material *mat, const aiMaterial *aimat)
{
	aiString name;
	aiColor4D aicol;
	float shin, shin_str;

	if(aiGetMaterialString(aimat, AI_MATKEY_NAME, &name) == 0) {
		mat->name = name.data;
	} else {
		mat->name = "unknown";
	}
	printf("load_material: %s\n", mat->name.c_str());

	if(aiGetMaterialColor(aimat, AI_MATKEY_COLOR_DIFFUSE, &aicol) == 0) {
		mat->diffuse = Vec3(aicol[0], aicol[1], aicol[2]);
	}
	if(aiGetMaterialColor(aimat, AI_MATKEY_COLOR_SPECULAR, &aicol) == 0) {
		mat->specular = Vec3(aicol[0], aicol[1], aicol[2]);
	}

	unsigned int count = 1;
	if(aiGetMaterialFloatArray(aimat, AI_MATKEY_SHININESS_STRENGTH, &shin_str, &count) != 0) {
		shin_str = 1.0;
	}
	if(aiGetMaterialFloatArray(aimat, AI_MATKEY_SHININESS, &shin, &count) == 0) {
		// XXX can't remember how I came up with this...
		mat->shininess = shin * shin_str * 0.0001 * 128.0;
	}

	// load textures

	const int num_tex_types = aiTextureType_UNKNOWN + 1;
	for(int i=0; i<num_tex_types; i++) {
		aiTextureType aitype = (aiTextureType)i;
		int count = aiGetMaterialTextureCount(aimat, aitype);

		for(int j=0; j<count; j++) {
			aiString aipath;
			if(aiGetMaterialTexture(aimat, aitype, j, &aipath) != 0) {
				continue;
			}

			char *fname;
			int nsize = datamap_path_size(aipath.data);
			if(nsize) {
				fname = new char[nsize];
				datamap_lookup(aipath.data, fname, nsize);
			} else {
				fname = new char[strlen(aipath.data) + 1];
				char *dptr = fname;
				char *sptr = aipath.data;
				do {
					*dptr++ = *sptr == '\\' ? '/' : *sptr;
				} while(*sptr++);
			}

			int textype = assimp_textype(aitype);
			printf("loading %s texture: %s\n", assimp_textypestr(aitype), fname);

			Texture *tex = scn->texset->get_texture(fname, TEX_2D);
			assert(tex);
			mat->textures.push_back(tex);

			printf("   DBG(%p)\n", (void*)tex);

			if(textype != MTL_TEX_UNKNOWN && !mat->stdtex[textype]) {
				mat->stdtex[textype] = tex;
			}
		}
	}

	return true;
}

static SceneNode *load_node(Scene *scn, const aiScene *aiscn, unsigned int flags, const aiNode *ainode)
{
	SceneNode *node = new SceneNode;
	node->set_name(ainode->mName.data);

	// transformation
	Mat4 matrix = assimp_matrix(ainode->mTransformation);
	Vec3 pos = matrix.get_translation();
	Quat rot = matrix.get_rotation();
	Vec3 scale = matrix.get_scaling();

	node->set_position(pos);
	node->set_rotation(rot);
	node->set_scaling(scale);
	node->dbg_xform = matrix;

	// meshes
	for(unsigned int i=0; i<ainode->mNumMeshes; i++) {
		aiMesh *aimesh = aiscn->mMeshes[ainode->mMeshes[i]];

		Mesh *mesh = mesh_by_aimesh[aimesh];
		if(mesh) {
			ObjMesh *obj = new ObjMesh;
			obj->mesh = mesh;
			// also grab the material of this mesh
			load_material(scn, &obj->mtl, aiscn->mMaterials[aimesh->mMaterialIndex]);

			node->add_object(obj);
			scn->objects.push_back(obj);
		}
	}

	/* recurse to all children */
	for(unsigned int i=0; i<ainode->mNumChildren; i++) {
		SceneNode *child = load_node(scn, aiscn, flags, ainode->mChildren[i]);
		if(child) {
			node->add_child(child);
		}
	}

	node_by_name[node->get_name()] = node;
	return node;
}

static Mesh *load_mesh(Scene *scn, const aiScene *aiscn, unsigned int flags, const aiMesh *aimesh)
{
	Mesh *mesh = new Mesh;

	int num_verts = aimesh->mNumVertices;
	int num_faces = aimesh->mNumFaces;

	mesh->set_attrib_data(MESH_ATTR_VERTEX, 3, num_verts, (float*)aimesh->mVertices);

	if(aimesh->mNormals) {
		mesh->set_attrib_data(MESH_ATTR_NORMAL, 3, num_verts, (float*)aimesh->mNormals);
	}
	if(aimesh->mTangents) {
		mesh->set_attrib_data(MESH_ATTR_TANGENT, 3, num_verts, (float*)aimesh->mTangents);
	}
	if(aimesh->mTextureCoords[0]) {
		mesh->set_attrib_data(MESH_ATTR_TEXCOORD, 3, num_verts, (float*)aimesh->mTextureCoords[0]);
	}
	if(aimesh->mTextureCoords[1]) {
		mesh->set_attrib_data(MESH_ATTR_TEXCOORD2, 3, num_verts, (float*)aimesh->mTextureCoords[1]);
	}

	if(flags & SCNLOAD_FLIPYZ) {
		Vec3 *vptr = (Vec3*)mesh->get_attrib_data(MESH_ATTR_VERTEX);
		for(int i=0; i<num_verts; i++) {
			*vptr = vptr->xzy();
			++vptr;
		}

		Vec3 *nptr = (Vec3*)mesh->get_attrib_data(MESH_ATTR_NORMAL);
		for(int i=0; i<num_verts; i++) {
			*nptr = nptr->xzy();
			++nptr;
		}

		Vec3 *tptr = (Vec3*)mesh->get_attrib_data(MESH_ATTR_TANGENT);
		for(int i=0; i<num_verts; i++) {
			*tptr = tptr->xzy();
			++tptr;
		}
	}

	unsigned int *iptr = mesh->set_index_data(num_faces * 3);
	for(int i=0; i<num_faces; i++) {
		iptr[0] = aimesh->mFaces[i].mIndices[0];
		iptr[1] = aimesh->mFaces[i].mIndices[flags & SCNLOAD_FLIPYZ ? 2 : 1];
		iptr[2] = aimesh->mFaces[i].mIndices[flags & SCNLOAD_FLIPYZ ? 1 : 2];
		iptr += 3;
	}

	AABox bbox = mesh->get_aabbox();
	printf("mesh bounding box: %f %f %f -> %f %f %f\n", bbox.min.x, bbox.min.y, bbox.min.z,
			bbox.max.x, bbox.max.y, bbox.max.z);

	return mesh;
}

static void print_nodes(SceneNode *node, int lvl)
{
	Vec3 pos = node->get_node_position();
	Quat rot = node->get_node_rotation();
	Vec3 scale = node->get_node_scaling();

	const char *type = node->get_num_objects() > 0 ? "mesh node" : "null node";

	for(int i=0; i<lvl; i++) {
		fputs("  ", stdout);
	}
	printf("%s[%s] p(%g %g %g) rq(%g %+gi %+gj %+gk) s(%g %g %g)\n", type, node->get_name(),
			pos.x, pos.y, pos.z, rot.w, rot.x, rot.y, rot.z, scale.x, scale.y, scale.z);

	if(node->get_num_objects()) {
		Mat4 xform = node->get_matrix();
		xform.print(stdout);
	}

	int nchld = node->get_num_children();
	for(int i=0; i<nchld; i++) {
		print_nodes(node->get_child(i), lvl + 1);
	}
}

static int assimp_textype(aiTextureType type)
{
	switch(type) {
	case aiTextureType_DIFFUSE:
		return MTL_TEX_DIFFUSE;
	case aiTextureType_SPECULAR:
		return MTL_TEX_SPECULAR;
	case aiTextureType_NORMALS:
		return MTL_TEX_NORMALMAP;
	case aiTextureType_LIGHTMAP:
	case aiTextureType_EMISSIVE:
		return MTL_TEX_LIGHTMAP;
	case aiTextureType_REFLECTION:
		return MTL_TEX_ENVMAP;
	default:
		break;
	}
	return MTL_TEX_UNKNOWN;
}

static const char *assimp_textypestr(aiTextureType type)
{
	switch(type) {
	case aiTextureType_DIFFUSE:
		return "diffuse";
	case aiTextureType_SPECULAR:
		return "specular";
	case aiTextureType_NORMALS:
		return "normalmap";
	case aiTextureType_LIGHTMAP:
	case aiTextureType_EMISSIVE:
		return "lightmap";
	case aiTextureType_REFLECTION:
		return "envmap";
	default:
		break;
	}
	return "unknown";
}

static Mat4 assimp_matrix(const aiMatrix4x4 &aim)
{
	Mat4 m;
	memcpy(m[0], &aim, 16 * sizeof(float));
	return transpose(m);
}
